/*
    xfmix - audio mixer
    Copyright (C) 1995  Radek Doulik

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <forms.h>
#include "light.h"
#include "xfmix.xpm"

#include <xpm.h>

#include <stdio.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#ifdef __linux
#include <linux/soundcard.h>
#else
#include <unistd.h>
#include <sys/soundcard.h>
#endif

FL_FORM  *mix_form, *about_form, *help_form;
FL_OBJECT *master_lock;

#define LEFT  (1<<24)
#define RIGHT (1<<25)
#define PAD   100

struct mixer {
  int handle;

  int devs;

  int devmask;
  int stereomask;
  int recmask;
  int recsrc;
} mix;

struct mix_dev {
  FL_OBJECT *l, *r;
  FL_OBJECT *fl_rec;
  FL_OBJECT *fl_mute;
  FL_OBJECT *fl_lock;

  int left, right;                                       /* values on L&R channels */
  int stereo;
  int dev_num;
  int dev_bit;
} *mdevs;

char *err_msg[]={ "Cannot open mixer",
		  "Error reading from mixer",
		  "No mixer devices found",
		  "Error writing to mixer",
		};

void end_error(int err) {
  fprintf(stderr, "%s\n", err_msg[err-1]);
  exit(err);
}

char *dev_name="/dev/mixer";
char *dev_labels[]=SOUND_DEVICE_LABELS;

void get_mix_info() {                                     /* open mixer and look for capabilities */

  if ((mix.handle=open(dev_name, O_RDWR))<0) end_error(1);

  if (ioctl(mix.handle, SOUND_MIXER_READ_DEVMASK, &mix.devmask)==-1) end_error(2);
  if (ioctl(mix.handle, SOUND_MIXER_READ_STEREODEVS, &mix.stereomask)==-1) end_error(2);
  if (ioctl(mix.handle, SOUND_MIXER_READ_RECMASK, &mix.recmask)==-1) end_error(2);

  if (!mix.devmask) end_error(3);
}

void set_volume(int dev_num, int left, int right) {
  unsigned char temp[4];

  temp[0]=(unsigned char)left;
  temp[1]=(unsigned char)right;
  temp[2]=temp[3]=0;
                                                          /* send it to mixer */
  ioctl(mix.handle, MIXER_WRITE(dev_num), temp);
}

void update_mixer_rec() {
  int i,j;

  if (ioctl(mix.handle, SOUND_MIXER_READ_RECSRC, &mix.recsrc)==-1) end_error(2);

  for (i=0; i<mix.devs; i++) {
    j=mdevs[i].dev_bit;
    if ((mix.devmask & j) && (mix.recmask & j))
      fl_set_lightbox(mdevs[i].fl_rec, mix.recsrc & j);
  }
}

void update_mixer_values() {
  int i;
  unsigned char temp[4];

  for (i=0; i<mix.devs; i++) {
    ioctl(mix.handle, MIXER_READ(mdevs[i].dev_num), &temp);
    mdevs[i].left=PAD*temp[0]/PAD;
    mdevs[i].right=PAD*temp[1]/PAD;
    fl_set_slider_value(mdevs[i].l, (double)temp[0]/PAD);
    fl_set_slider_value(mdevs[i].r, (double)temp[1]/PAD);
  }
}

void lock_dev(int i, int ns) {
  if (ns) {
    int m=fl_get_lightbox(mdevs[i].fl_mute);

    mdevs[i].left=mdevs[i].right=(mdevs[i].left+mdevs[i].right)/2;
    set_volume(mdevs[i].dev_num, mdevs[i].left, mdevs[i].right);
    fl_set_slider_value(mdevs[i].l, (m) ? 0 : (double)mdevs[i].left/PAD);
    fl_set_slider_value(mdevs[i].r, (m) ? 0 : (double)mdevs[i].right/PAD);
  }
  fl_set_lightbox(mdevs[i].fl_lock, ns);
  fl_set_object_color(mdevs[i].l, FL_COL1, (ns) ? FL_INACTIVE : FL_RED);
  fl_set_object_color(mdevs[i].r, FL_COL1, (ns) ? FL_INACTIVE : FL_BLUE);
}

void mix_master_lock_cb(FL_OBJECT *ob, long l) {          /* callback for master lock */
  int ns=fl_get_lightbox(ob), i;

  for (i=0;i<mix.devs;i++) {
    if (fl_get_lightbox(mdevs[i].fl_lock)!=ns)
      lock_dev(i, ns);
  }
}

void mix_master_mute_cb(FL_OBJECT *ob, long l) {          /* callback for master mute */
  int st=1-fl_get_lightbox(ob);
  fl_set_lightbox(ob, st);
  set_volume(SOUND_MIXER_MUTE, st, 0);
}

void mix_mute_cb(FL_OBJECT *ob, long data) {              /* callback for channel mute */
  int st=fl_get_lightbox(ob);
  int l, r;
  
  if (st) set_volume(mdevs[data].dev_num, l=0, r=0); 
  else set_volume(mdevs[data].dev_num, l=mdevs[data].left, r=mdevs[data].right);

  fl_set_slider_value(mdevs[data].l, (double)l/PAD);
  fl_set_slider_value(mdevs[data].r, (double)r/PAD);
}

void mix_slider_cb(FL_OBJECT *ob, long data) {            /* callback for mixer sliders */
  int vol, idx=data & 255;
  double v;

  v=fl_get_slider_value(ob);                              /* get current volume */
  vol=(int)(PAD*v);
  if (data & LEFT) mdevs[idx].left=vol;
  else mdevs[idx].right=vol;

  if (fl_get_lightbox(mdevs[idx].fl_lock)) {              /* locking feature */
    if (data & LEFT) {
      mdevs[idx].right=vol;
      fl_set_slider_value(mdevs[idx].r, v);
    } else {
      mdevs[idx].left=vol;
      fl_set_slider_value(mdevs[idx].l, v);
    }
  }

  set_volume(mdevs[idx].dev_num, mdevs[idx].left, mdevs[idx].right);
  if ((mdevs[idx].left>0 || mdevs[idx].right>0) && fl_get_lightbox(mdevs[idx].fl_mute))
    fl_set_lightbox(mdevs[idx].fl_mute, 0);
}

void mix_lock_cb(FL_OBJECT *ob, long data) {             /* callback for locking */
  int ns=fl_get_lightbox(ob);

  lock_dev(data, ns);
  if (!ns && fl_get_lightbox(master_lock))
    fl_set_lightbox(master_lock, 0);
}

void mix_rec_cb(FL_OBJECT *ob, long data) {              /* callback for recording */

  if (fl_get_lightbox(ob)) {
    mix.recsrc|=mdevs[data].dev_bit;
  } else mix.recsrc&=(0xffffffff ^ mdevs[data].dev_bit);

  if (ioctl(mix.handle, SOUND_MIXER_WRITE_RECSRC, &mix.recsrc)==-1) end_error(4);

  update_mixer_rec();
}

void mix_end_cb(FL_OBJECT *o, long l) {
  exit(0);
}

void about_end_cb(FL_OBJECT *o, long l) {
  FL_FORM *f=(FL_FORM*)l;

  fl_hide_form(f);
  fl_free_form(f);
  fl_activate_all_forms();
}

void mix_help_cb(FL_OBJECT *o, long l) {
  FL_OBJECT *ob;

  fl_deactivate_all_forms();

  help_form = fl_bgn_form(FL_UP_BOX, 260, 125);

  fl_add_text(FL_NORMAL_TEXT,20,30,80,60,
	      "use 'l' to lock / unlock all channels\n"
	      "      'm' to mute/unmute master (if supported)");
  fl_add_frame(FL_DOWN_FRAME,10,40,240,75,"");
  ob=fl_add_text(FL_NORMAL_TEXT,10,75,65,40,"help");
    fl_set_object_lsize(ob,FL_HUGE_SIZE);
    fl_set_object_lstyle(ob,FL_NORMAL_STYLE+FL_SHADOW_STYLE);
  ob=fl_add_text(FL_NORMAL_TEXT,185,75,65,40,"???");
    fl_set_object_lsize(ob,FL_HUGE_SIZE);
    fl_set_object_lstyle(ob,FL_NORMAL_STYLE+FL_SHADOW_STYLE);
  ob=fl_add_button(FL_NORMAL_BUTTON,185,10,65,20,"Close");
    fl_set_object_callback(ob, about_end_cb, (long)help_form);
  fl_end_form();

  fl_show_form(help_form, FL_PLACE_MOUSE, FL_TRANSIENT, "xfmix help");
}

void mix_about_cb(FL_OBJECT *o, long l) {
  FL_OBJECT *ob;

  fl_deactivate_all_forms();

  about_form = fl_bgn_form(FL_UP_BOX, 220, 140);

  fl_add_box(FL_UP_BOX,0,0,220,140,"");
  fl_add_frame(FL_DOWN_FRAME,10,40,200,90,"");
  fl_add_text(FL_NORMAL_TEXT,10,50,200,35,
	      "xforms based mixer      for Linux/X11R6\n\n"
	      "(c) 1995 R.Doulik     rodo@earn.cvut.cz");
  ob=fl_add_text(FL_NORMAL_TEXT,10,95,90,30,"xfmix");
    fl_set_object_lsize(ob, FL_HUGE_SIZE);
    fl_set_object_lstyle(ob, FL_NORMAL_STYLE+FL_SHADOW_STYLE);
  fl_add_text(FL_NORMAL_TEXT,105,95,100,20,"version 0.1 ALPHA");
  ob=fl_add_button(FL_NORMAL_BUTTON,145,10,65,20,"Close");
    fl_set_object_callback(ob, about_end_cb, (long)about_form);

  fl_end_form();

  fl_show_form(about_form, FL_PLACE_MOUSE, FL_TRANSIENT, "xfmix about");
}

void create_form_mix() {
  FL_OBJECT* ob;
  unsigned int x,i,j,k;

  get_mix_info();

  for (x=i=mix.devs=0, j=1; i<20; i++, j<<=1)             /* count form width */
    if (mix.devmask & j) {
      x+=(mix.stereomask & j) ? 45 : 30;
      mix.devs++;
    }

  mdevs=(struct mix_dev*)malloc(sizeof(struct mix_dev)*mix.devs);

  fl_set_border_width(-2);
  mix_form=fl_bgn_form(FL_FLAT_BOX, x+75, 260);

  x+=15;
  ob=fl_add_button(FL_NORMAL_BUTTON, x, 53, 50, 20, "Quit");
    fl_set_object_callback(ob, mix_end_cb, 0);
  ob=fl_add_button(FL_NORMAL_BUTTON, x, 232, 50, 20, "Help");
    fl_set_object_callback(ob, mix_help_cb, 0);
  ob=fl_add_button(FL_NORMAL_BUTTON, x, 207, 50, 20, "About");
    fl_set_object_callback(ob, mix_about_cb, 0);

  if (mix.devmask & (1<<SOUND_MIXER_MUTE)) {
    ob=fl_add_lightbox(FL_NORMAL_LIGHTBOX, x+5, 25, 45, 8, "mute");
      fl_set_object_callback(ob, mix_master_mute_cb, 0);
      fl_set_object_shortcut(ob, "mM", 0);
  }

  master_lock=ob=fl_add_lightbox(FL_NORMAL_LIGHTBOX, x+5, 10, 45, 8, "lock");
    fl_set_object_color(ob, FL_COL1, FL_WHITE); 
    fl_set_object_callback(ob, mix_master_lock_cb, 0);
    fl_set_object_shortcut(ob, "lL", 0);

  for (i=k=0, j=1, x=10; i<20; i++, j<<=1) {
    if (mix.devmask & j) {
      mdevs[k].stereo=0;
      mdevs[k].dev_num=i;
      mdevs[k].dev_bit=j;

      fl_add_text(FL_NORMAL_TEXT, x, 52, 40, 20, dev_labels[i]);
      fl_add_frame(FL_DOWN_FRAME,x,55, (mix.stereomask & j) ? 40 : 25, 195,"");

      if (mix.recmask & j) {
	mdevs[k].fl_rec=ob=fl_add_lightbox(FL_NORMAL_LIGHTBOX, x, 40, 45, 8, "rec");
	  fl_set_object_color(ob, FL_COL1, FL_RED); 
          fl_set_object_callback(ob, mix_rec_cb, k);
      }
      mdevs[k].fl_mute=ob=fl_add_lightbox(FL_NORMAL_LIGHTBOX, x, 25, 45, 8, "mute");
        fl_set_object_callback(ob, mix_mute_cb, k);

      mdevs[k].fl_lock=ob=fl_add_lightbox(FL_NORMAL_LIGHTBOX, x, 10, 45, 8, "lock");
        fl_set_object_callback(ob, mix_lock_cb, k);
        fl_set_object_color(ob, FL_COL1, FL_WHITE); 

      mdevs[k].l=ob=fl_add_valslider(FL_VERT_NICE_SLIDER,x,70,20,180, "");
        fl_set_object_boxtype(ob, FL_FLAT_BOX);
        fl_set_object_color(ob, FL_COL1,
			    (fl_get_lightbox(mdevs[k].fl_lock)) ? FL_INACTIVE : FL_RED);
        fl_set_object_callback(ob, mix_slider_cb, k | LEFT);
        fl_set_slider_bounds(ob, 1, 0);

      if (mix.stereomask & j) {
	mdevs[k].stereo=1;
	mdevs[k].r=ob=fl_add_valslider(FL_VERT_NICE_SLIDER,x+20,70,20,180,"");
	  fl_set_object_boxtype(ob, FL_FLAT_BOX);
	  fl_set_object_color(ob, FL_COL1,
			      (fl_get_lightbox(mdevs[k].fl_lock)) ? FL_INACTIVE : FL_BLUE);
          fl_set_object_callback(ob, mix_slider_cb, k | RIGHT);
          fl_set_slider_bounds(ob, 1, 0);
	x+=15;
      }
      x+=30;
      k++;
    }
  }

  fl_end_form();

  update_mixer_values();
  update_mixer_rec();
}

int main(int argc, char *argv[]) {
  Pixmap icon, icon_shape;
  XWMHints *wm_hints;

  /* initialize forms library */
  fl_initialize(argv[0], "LandScape", 0, 0, &argc, argv);
  
  /* create main window form */   
  create_form_mix();
  
  /* now we can begin */
  fl_show_form(mix_form, FL_PLACE_FREE, FL_FULLBORDER, "Mixer");

  /* set my icon */
  XpmCreatePixmapFromData(fl_get_display(), mix_form->window, xfmix_icon, &icon, &icon_shape,
			  NULL);

  wm_hints=XAllocWMHints();
  wm_hints->flags=IconPixmapHint;
  wm_hints->icon_pixmap=icon;
  XSetWMHints(fl_get_display(), mix_form->window, wm_hints);

  while(1) fl_do_forms();         /* do it forever :) */

  XFree(wm_hints);
}
